/*************************************************************************
  vStrip by [maven] (maven@maven.de)
	parse_ifo.c: Parse the Cell-Position from the ProgramChain (PGC)
							 in an IFO-file. (tabsize 2)

  based on code originally written by Thomas Mirlacher (dent@cosy.sbg.ac.at)
  for the "free InFormatiOn project" (http://www.linuxvideo.org)

  Many Thanks!
*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef __UNIX__
#include <io.h>
#else
#include <unistd.h>
#endif
#include <fcntl.h>
#include "s_types.h"
#include "parse_ifo.h"
#include "file_io.h"
#include "vstrip.h"

#define get2bytes(b) bswap_16(*((word *)(b)))
#define get4bytes(b) bswap_32(*((dword *)(b)))
#define be2me_16(x) bswap_16(x)
#define be2me_32(x) bswap_32(x)

static struct
{
	char descr[3];
  char lang_long[20];
} iifo_lang_tbl[] = {
/* The ISO 639 language codes.
 */
	{"  ", "Not Specified"},
	{"aa", "Afar"},
	{"ab", "Abkhazian"},
	{"af", "Afrikaans"},
	{"am", "Amharic"},
	{"ar", "Arabic"},
	{"as", "Assamese"},
	{"ay", "Aymara"},
	{"az", "Azerbaijani"},
	{"ba", "Bashkir"},
	{"be", "Byelorussian"},
	{"bg", "Bulgarian"},
	{"bh", "Bihari"},
	{"bi", "Bislama"},
	{"bn", "Bengali; Bangla"},
	{"bo", "Tibetan"},
	{"br", "Breton"},
	{"ca", "Catalan"},
	{"co", "Corsican"},
	{"cs", "Czech(Ceske)"},
	{"cy", "Welsh"},
	{"da", "Dansk"},
	{"de", "Deutsch"},
	{"dz", "Bhutani"},
	{"el", "Greek"},
	{"en", "English"},
	{"eo", "Esperanto"},
	{"es", "Espanol"},
	{"et", "Estonian"},
	{"eu", "Basque"},
	{"fa", "Persian"},
	{"fi", "Suomi"},
	{"fj", "Fiji"},
	{"fo", "Faroese"},
	{"fr", "Francais"},
	{"fy", "Frisian"},
	{"ga", "Irish"},
	{"gd", "Scots Gaelic"},
	{"gl", "Galician"},
	{"gn", "Guarani"},
	{"gu", "Gujarati"},
	{"ha", "Hausa"},
	{"he", "Hebrew"},				// formerly iw
	{"hi", "Hindi"},
	{"hr", "Hrvatski"},				// Croatian
	{"hu", "Magyar"},
	{"hy", "Armenian"},
	{"ia", "Interlingua"},
	{"id", "Indonesian"},				// formerly in
	{"ie", "Interlingue"},
	{"ik", "Inupiak"},
	{"in", "Indonesian"},				// replaced by id
	{"is", "Islenska"},
	{"it", "Italiano"},
	{"iu", "Inuktitut"},
	{"iw", "Hebrew"},				// replaced by he
	{"ja", "Japanese"},
	{"ji", "Yiddish"},				// replaced by yi
	{"jw", "Javanese"},
	{"ka", "Georgian"},
	{"kk", "Kazakh"},
	{"kl", "Greenlandic"},
	{"km", "Cambodian"},
	{"kn", "Kannada"},
	{"ko", "Korean"},
	{"ks", "Kashmiri"},
	{"ku", "Kurdish"},
	{"ky", "Kirghiz"},
	{"la", "Latin"},
	{"ln", "Lingala"},
	{"lo", "Laothian"},
	{"lt", "Lithuanian"},
	{"lv", "Latvian, Lettish"},
	{"mg", "Malagasy"},
	{"mi", "Maori"},
	{"mk", "Macedonian"},
	{"ml", "Malayalam"},
	{"mn", "Mongolian"},
	{"mo", "Moldavian"},
	{"mr", "Marathi"},
	{"ms", "Malay"},
	{"mt", "Maltese"},
	{"my", "Burmese"},
	{"na", "Nauru"},
	{"ne", "Nepali"},
	{"nl", "Nederlands"},
	{"no", "Norsk"},
	{"oc", "Occitan"},
	{"om", "(Afan) Oromo"},
	{"or", "Oriya"},
	{"pa", "Punjabi"},
	{"pl", "Polish"},
	{"ps", "Pashto, Pushto"},
	{"pt", "Portugues"},
	{"qu", "Quechua"},
	{"rm", "Rhaeto-Romance"},
	{"rn", "Kirundi"},
	{"ro", "Romanian"},
	{"ru", "Russian"},
	{"rw", "Kinyarwanda"},
	{"sa", "Sanskrit"},
	{"sd", "Sindhi"},
	{"sg", "Sangho"},
	{"sh", "Serbo-Croatian"},
	{"si", "Sinhalese"},
	{"sk", "Slovak"},
	{"sl", "Slovenian"},
	{"sm", "Samoan"},
	{"sn", "Shona"},
	{"so", "Somali"},
	{"sq", "Albanian"},
	{"sr", "Serbian"},
	{"ss", "Siswati"},
	{"st", "Sesotho"},
	{"su", "Sundanese"},
	{"sv", "Svenska"},
	{"sw", "Swahili"},
	{"ta", "Tamil"},
	{"te", "Telugu"},
	{"tg", "Tajik"},
	{"th", "Thai"},
	{"ti", "Tigrinya"},
	{"tk", "Turkmen"},
	{"tl", "Tagalog"},
	{"tn", "Setswana"},
	{"to", "Tonga"},
	{"tr", "Turkish"},
	{"ts", "Tsonga"},
	{"tt", "Tatar"},
	{"tw", "Twi"},
	{"ug", "Uighur"},
	{"uk", "Ukrainian"},
	{"ur", "Urdu"},
	{"uz", "Uzbek"},
	{"vi", "Vietnamese"},
	{"vo", "Volapuk"},
	{"wo", "Wolof"},
	{"xh", "Xhosa"},
	{"yi", "Yiddish"},				// formerly ji
	{"yo", "Yoruba"},
	{"za", "Zhuang"},
	{"zh", "Chinese"},
	{"zu", "Zulu"},
	{"\0", "???"}
};

static char *iifoDecodeLang(uint16_t descr)
{
	int i;

	for (i = 0; iifo_lang_tbl[i].lang_long[0]; i++)
		if (!memcmp(&descr, &iifo_lang_tbl[i].descr, 2))
			return iifo_lang_tbl[i].lang_long;
	return iifo_lang_tbl[i].lang_long;
}

static char *iifoDecodeAudioMode(uint16_t descr)
{
	static char audio_coding_mode[][9] = {
		"AC3",
		"???",
		"MPEG1",
		"MPEG2ext",
		"LPCM",
		"DTS",
		"SDDS"
	};

	return audio_coding_mode[descr & 0x07];
}

static char *iifoDecodeAudioStreamID(uint16_t descr, const int audio_idx)
{
	static dword stream_ofs[] = {0xbd80, 0x0000, 0xc000, 0xc800, 0xbda0, 0xbd88, 0xbd88};
	static char stream_id[10];

	if ((stream_ofs[descr & 7] & 0xff) != 0)
		sprintf(stream_id, "0x%02X 0x%02X", stream_ofs[descr & 7] >> 8, audio_idx + (stream_ofs[descr & 7] & 0xff));
	else
		sprintf(stream_id, "0x%02X", audio_idx + (stream_ofs[descr & 7] >> 8));
	return stream_id;
}

const ifo_pgci_caddr_t *iifoGetCellPlayInfo(const pgc_t *pgc)
{
	const uint8_t *ptr = (uint8_t *) pgc;

	return (ifo_pgci_caddr_t *)(ptr + be2me_16(pgc->cell_playback_tbl_offset));
}

uint8_t iifoGetCellPosNum(const pgc_t *pgc)
{
	return pgc->nr_of_cells;
}

const ifo_pgc_cpos_t *iifoGetCellPos(const pgc_t *pgc)
{
	const uint8_t *ptr = (uint8_t *)pgc;

	return (ifo_pgc_cpos_t *)(ptr + be2me_16(pgc->cell_position_tbl_offset));
}

// ---- Toff start ----
uint8_t iifoGetProgramMapNum(const pgc_t *pgc)
{
	return pgc->nr_of_programs;
}

const pgc_program_map_t *iifoGetProgramMap(const pgc_t *pgc)
{
	const uint8_t *ptr = (uint8_t *)pgc;

	return (pgc_program_map_t *)(ptr + be2me_16(pgc->pgc_program_map_offset));
}
// ---- Toff end ----

const pgc_t *iifoGetPGCI(const ifo_t *ifo, const int title)
{
	pgci_sub_t *pgci_sub;
	ifo_hdr_t *hdr = NULL;
	uint8_t *ptr = NULL;

	hdr = (ifo_hdr_t *)ifo->title_pgci;

	if (hdr && title < bswap_16(hdr->num))
	{
		ptr = (uint8_t *)hdr;
		ptr += IFO_HDR_LEN;

		pgci_sub = (pgci_sub_t *)ptr + title;

		ptr = (uint8_t *)hdr + be2me_32(pgci_sub->start);
	}

	return (const pgc_t *)ptr;
}

#define OFF_TITLE_PGCI get4bytes (ifo->mat + 0xCC)

bool iifoIsVTS(ifo_t *ifo)
{
	return strncmp (ifo->mat, "DVDVIDEO-VTS", 12) == 0;
}

int iifoReadLB(tp_fio_file fp, dword pos, dword count, byte *data)
{
	if ((pos = fio_direct_lseek(fp, pos, SEEK_SET)) == -1)
    return -1;
	return fio_direct_read(fp, data, count);
}

bool iifoReadTBL(tp_fio_file fp, dword offset, byte **out_ptr)
{
	byte *data;
	ifo_hdr_t *hdr;
	dword len = 0;

	if (!offset)
		return FALSE;

	data =  (byte *)malloc(fio_SECTOR_SIZE);
	if (!data)
		return FALSE;

	if (iifoReadLB (fp, offset * fio_SECTOR_SIZE, fio_SECTOR_SIZE, data) <= 0)
		return FALSE;

	hdr = (ifo_hdr_t *)data;
	len = bswap_32 (hdr->len) + 1;

	if (len > fio_SECTOR_SIZE)
	{
		if (!(data = (byte *)realloc((void *)data, len)))
			return FALSE;
		iifoReadLB (fp, offset * fio_SECTOR_SIZE, len, data);
	}
	*out_ptr = data;
	return TRUE;
}

// interface

ifo_t *ifoOpen(const char *name, const dword fio_flags)
{
	ifo_t *ret_ifo = NULL;
	tp_fio_file fd;

	fd = fio_open(name, fio_flags, 0);
	if (fd)
	{
		ifo_t *ifo = calloc(sizeof *ifo, 1);

		ifo->mat = calloc(fio_SECTOR_SIZE, 1);
		if (iifoReadLB(fd, 0, fio_SECTOR_SIZE, ifo->mat) < 0)
		{
			free(ifo->mat);
			free(ifo);
		}
		else 
		{ // looking ok so far
			if (iifoIsVTS(ifo) && iifoReadTBL(fd, OFF_TITLE_PGCI, &ifo->title_pgci))
				ret_ifo = ifo;
			else
			{
				if (ifo->title_pgci)
					free(ifo->title_pgci);
				free (ifo->mat);
				free (ifo);
			}
		}
		fio_close(fd);
	}
	return ret_ifo;
}

bool ifoClose (ifo_t *ifo)
{
	if (ifo)
	{
		if (ifo->mat)
			free (ifo->mat);
		if (ifo->title_pgci)
			free (ifo->title_pgci);
		free (ifo);
	}
	return TRUE;
}

int ifoGetNumPGCI(const ifo_t *ifo)
{
	int num = 0;

	if (ifo && ifo->title_pgci)
		num = bswap_16(((ifo_hdr_t *)ifo->title_pgci)->num);
	return num;
}

// returns the number of cells and the length in time_out
int ifoGetPGCIInfo(const ifo_t *ifo, const dword title, byte time_out[4])
{
	int num_cells = 0;
	const pgc_t *pgc;

	if (ifo && (pgc = iifoGetPGCI(ifo, title)))
	{
		num_cells = pgc->nr_of_cells;
		if (time_out)
		{
			int i;

			time_out[0] = pgc->playback_time.hour;
			time_out[1] = pgc->playback_time.minute;
			time_out[2] = pgc->playback_time.second;
			time_out[3] = pgc->playback_time.frame_u & DVD_TIME_AND;
			for (i = 0; i < 4; i++)
				time_out[i] = ((time_out[i] & 0xf0) >> 4) * 10 + (time_out[i] & 0x0f);
		}
	}
	return num_cells;
}

// ---- Toff start ----
int ifoGetProgramMapNum(const ifo_t *ifo, const dword title) {
  const pgc_t *pgc;
  if( (pgc = iifoGetPGCI(ifo, title)) ) {
    return pgc->nr_of_programs;
  }
  return -1;
}

int ifoGetProgramMap(const ifo_t *ifo, const dword title, pgc_program_map_t* cellsPM)
{
  bool num_prog = 0;
  const pgc_t *pgc;
  const pgc_program_map_t *pm;
  
  if (ifo && cellsPM && (pgc = iifoGetPGCI(ifo, title)) && (pm = iifoGetProgramMap(pgc))) {
      dword i;

		  num_prog = pgc->nr_of_programs;
      for (i = 0; i < pgc->nr_of_programs; i++) {
        cellsPM[i] = pm[i];
      }      
  }

  return num_prog;
}
// ---- Toff end ----

bool ifoGetPGCICells(const ifo_t *ifo, const dword title, tp_vs_vobcellid cells)
{
	bool success = FALSE;
	const pgc_t *pgc;
	const ifo_pgc_cpos_t *cpos;
	const ifo_pgci_caddr_t *cpi;

	if (ifo && cells && (pgc = iifoGetPGCI(ifo, title)) && (cpos = iifoGetCellPos(pgc)) && (cpi = iifoGetCellPlayInfo(pgc)))
	{
		dword i, j, k = 0, anum = 0;

		success = TRUE;
		for (i = 0; i < pgc->nr_of_cells; i++)
		{
			cells[i].start_lba = bswap_32(cpi[i].vobu_start);
			cells[i].end_lba = bswap_32(cpi[i].vobu_last_end);
			cells[i].vob_id = bswap_16(cpos[i].vob_id);
			cells[i].cell_id = cpos[i].cell_id;

      // ---- Toff start ----
			cells[i].time_out[0] = cpi[i].playback_time.hour;
			cells[i].time_out[1] = cpi[i].playback_time.minute;
			cells[i].time_out[2] = cpi[i].playback_time.second;
			cells[i].time_out[3] = cpi[i].playback_time.frame_u & DVD_TIME_AND;
			for (j = 0; j < 4; j++)
				cells[i].time_out[j] = ((cells[i].time_out[j] & 0xf0) >> 4) * 10 + (cells[i].time_out[j] & 0x0f);
      // ---- Toff end ----

			switch (cpi[i].chain_info & 0xc0)
			{
				case 0x00: // 00
					cells[i].angle = 0x11; // 1 of 1
					k = i;
				break;
				case 0x40: // 01
					cells[i].angle = 0xf1; // 1 of ?
					k = i;
					anum = 1;
				break;
				case 0x80: // 10
					cells[i].angle = 0xf0 | (++anum & 0xf); // anum of ?
				break;
				case 0xc0: // 11
					cells[i].angle = 0xf0 | (++anum & 0xf); // anum of ?
					anum = (anum << 4) & 0xf0;
					for (j = k; j <= i; j++) // now set all the ones we've found
						if ((cells[j].angle & 0xf0) == 0xf0)
							cells[j].angle = (cells[j].angle & 0x0f) | anum;
				break;
			}
		}
	}
	return success;
}

int ifoGetNumAudio(const ifo_t *ifo)
{
	int num = 0;
	
	if (ifo && ifo->mat)
	{
		audio_hdr_t *hdr = (audio_hdr_t *)(ifo->mat + IFO_OFFSET_AUDIO);
		bool n[8];
		int i, j, k;
		
		memset(n, 0, sizeof n);
		j = ifoGetNumPGCI(ifo);
		for (i = 0; i < j; i++)
		{
			const pgc_t *pgc = iifoGetPGCI(ifo, i);

			if (pgc)
			{
				for (k = 0; k < 8; k++)
					if (pgc->audio_status[k].available)
						n[k] = TRUE;
			}
		}
		for (i = 0; i < 8; i++)
			if (n[i])
				num++;

		num = min(num, bswap_16(hdr->num));
	}
	return num;
}

char *ifoGetAudioDesc(const ifo_t *ifo, const int audio_idx)
{
	static char out[256];

	if (ifo && ifo->mat)
	{
		audio_hdr_t *hdr = (audio_hdr_t *)(ifo->mat + IFO_OFFSET_AUDIO);

		if (audio_idx < ifoGetNumAudio(ifo))
		{
			ifo_audio_t *audio = (ifo_audio_t *)((byte *)hdr + AUDIO_HDR_LEN + audio_idx * IFO_AUDIO_LEN);

			sprintf(out, "%s (%s %dch, %s)", iifoDecodeLang(audio->lang_code), iifoDecodeAudioMode(audio->coding_mode), audio->num_channels + 1, iifoDecodeAudioStreamID(audio->coding_mode, audio_idx));
			return out;
		}
	}
	return NULL;
}

int ifoGetNumSubPic(const ifo_t *ifo)
{
	int num = 0;
	
	if (ifo && ifo->mat)
	{
		spu_hdr_t *hdr = (spu_hdr_t *)(ifo->mat + IFO_OFFSET_SUBPIC);
		bool n[32];
		int i, j, k;
		
		memset(n, 0, sizeof n);
		j = ifoGetNumPGCI(ifo);
		for (i = 0; i < j; i++)
		{
			const pgc_t *pgc = iifoGetPGCI(ifo, i);

			if (pgc)
			{
				for (k = 0; k < 32; k++)
					if (pgc->subp_status[k].available)
						n[k] = TRUE;
			}
		}
		for (i = 0; i < 32; i++)
			if (n[i])
				num++;

		num = min(num, hdr->num);
	}
	return num;
}

char *ifoGetSubPicDesc(const ifo_t *ifo, const int subp_idx)
{
	static char out[256];

	if (ifo && ifo->mat)
	{
		spu_hdr_t *hdr = (spu_hdr_t *)(ifo->mat + IFO_OFFSET_SUBPIC);

		if (subp_idx < hdr->num)
		{
			ifo_spu_t *sub = (ifo_spu_t *)((byte *)hdr + SPU_HDR_LEN + subp_idx * sizeof *sub);

			sprintf(out, "%s (0xBD 0x%02X)", iifoDecodeLang(sub->lang_code), 0x20 + subp_idx);
			return out;
		}
	}
	return NULL;
}

bool ifoParse(const char *ifo_name, const int ifo_pgc_idx, const t_ifo_flags flags, dword *num_cells, tp_vs_vobcellid *cell_ids)
{
	if (ifo_name && *ifo_name)
	{
		ifo_t *ifo = NULL;
		t_fio_flags ffl = 0;
		int i, j;

		if ((flags & ifo_PRINT_INFO) != 0)
			fprintf(stdout, "Parsing \"%s\"...\n", ifo_name);
		
		if ((flags & ifo_USE_ASPI) != 0)
			ffl |= fio_USE_ASPI;
		if ((flags & ifo_PREFER_ASPI) != 0)
			ffl |= fio_PREFER_ASPI;
		ifo = ifoOpen(ifo_name, ffl);
		if (ifo)
		{
			int num = ifoGetNumPGCI(ifo);

			if ((flags & ifo_PRINT_INFO) != 0)
			{
				j = ifoGetNumAudio(ifo);
				if (j > 0)
					fputs("Audio:\n", stdout);
				for (i = 0; i < j; i++)
					fprintf(stdout, "  %u. %s\n", i, ifoGetAudioDesc(ifo, i));
				
				j = ifoGetNumSubPic(ifo);
				if (j > 0)
					fputs("SubPicture:\n", stdout);
				for (i = 0; i < j; i++)
					fprintf(stdout, "  %u. %s\n", i, ifoGetSubPicDesc(ifo, i));
				
				fputs("Program Chain(s):\n", stdout);
			}

			if (ifo_pgc_idx >= 0)
			{
				if (ifo_pgc_idx < num)
				{
					byte time[4];

					*num_cells = ifoGetPGCIInfo(ifo, ifo_pgc_idx, time);
					*cell_ids = calloc(*num_cells, sizeof **cell_ids);
					if ((flags & ifo_PRINT_INFO) != 0)
						fprintf(stdout, "  Using PGC #%u. Length: %02u:%02u:%02u:%02u in %i cell(s).\n", ifo_pgc_idx, time[0], time[1], time[2], time[3], *num_cells);
					ifoGetPGCICells(ifo, ifo_pgc_idx, *cell_ids);

					if ((flags & ifo_PRINT_CELL_POS) != 0 && (flags & ifo_PRINT_INFO) != 0)
						for (j = 0; j < (int)*num_cells; j++)
						{
							if ((*cell_ids)[j].angle != 0x11)
								fprintf(stdout, "    %u. VOB-ID: %02u/CELL-ID: %02u (Angle %01u/%01u).\n", j, (*cell_ids)[j].vob_id, (*cell_ids)[j].cell_id, ((*cell_ids)[j].angle) & 0xf, ((*cell_ids)[j].angle) >> 4);
							else
								fprintf(stdout, "    %u. VOB-ID: %02u/CELL-ID: %02u.\n", j, (*cell_ids)[j].vob_id, (*cell_ids)[j].cell_id);
						}
				}
				else
				{
					ifoClose(ifo);
					if ((flags & ifo_PRINT_INFO) != 0)
			      fprintf(stderr, "* Program-Chain (PGC) Index out of range (only %i PGC(s) available)\n", num);
			    return FALSE;
				}
			}
			else
			{ // print all PGCs and select none
				for (i = 0; i < num; i++)
				{
					byte time[4];
					int n_cells;

					n_cells = ifoGetPGCIInfo(ifo, i, time);
					
					if ((flags & ifo_PRINT_INFO) != 0)
						fprintf(stdout, "  %u. Length: %02u:%02u:%02u:%02u in %i cell(s).\n", i, time[0], time[1], time[2], time[3], n_cells);
					if (n_cells > 0 && (flags & ifo_PRINT_CELL_POS) != 0 && (flags & ifo_PRINT_INFO) != 0)
					{
						t_vs_vobcellid *cellids;

						cellids = calloc(n_cells, sizeof *cellids);
						ifoGetPGCICells(ifo, i, cellids);
						for (j = 0; j < n_cells; j++)
						{
							if (cellids[j].angle != 0x11)
								fprintf(stdout, "    %u. VOB-ID: %02u/CELL-ID: %02u (Angle %01u/%01u).\n", j, cellids[j].vob_id, cellids[j].cell_id, cellids[j].angle & 0xf, cellids[j].angle >> 4);
							else
								fprintf(stdout, "    %u. VOB-ID: %02u/CELL-ID: %02u.\n", j, cellids[j].vob_id, cellids[j].cell_id);
						}
						free(cellids);
					}
				}
			}
			ifoClose(ifo);
		}
		else
		{
			if ((flags & ifo_PRINT_INFO) != 0)
	      fprintf(stderr, "* Unable to parse \"%s\" as IFO-File\n", ifo_name);
	    return FALSE;
		}
	}
	return TRUE;
}